#!/bin/bash
###############################################################################
#     parabola-vmbootstrap -- create and start parabola virtual machines      #
#                                                                             #
#     Copyright (C) 2017 - 2019  Andreas Grapentin                            #
#                                                                             #
#     This program is free software: you can redistribute it and/or modify    #
#     it under the terms of the GNU General Public License as published by    #
#     the Free Software Foundation, either version 3 of the License, or       #
#     (at your option) any later version.                                     #
#                                                                             #
#     This program is distributed in the hope that it will be useful,         #
#     but WITHOUT ANY WARRANTY; without even the implied warranty of          #
#     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           #
#     GNU General Public License for more details.                            #
#                                                                             #
#     You should have received a copy of the GNU General Public License       #
#     along with this program.  If not, see <http://www.gnu.org/licenses/>.   #
###############################################################################


# shellcheck source=/usr/lib/libretools/messages.sh
source "$(librelib messages)"


# defaults
readonly DEF_KERNEL='linux-libre'
readonly DEF_MIRROR="https://repo.parabola.nu/\$repo/os/\$arch"
readonly DEF_IMG_GB=64
readonly MIN_GB=1
readonly DEF_BOOT_MB=100
readonly DEF_SWAP_MB=0

# misc
readonly THIS_DIR=$(dirname ${BASH_SOURCE})

# options
Hooks=()
Kernel=$DEF_KERNEL
Mirror=$DEF_MIRROR
IsNonsystemd=0
ImgSizeGb=$DEF_IMG_GB
BootSizeMb=$DEF_BOOT_MB
SwapSizeMb=$DEF_SWAP_MB


usage() {
  print "USAGE:"
  print "  pvmbootstrap [-h] [-H <hook>] [ -k <kernel> ] [-M <mirror>]"
  print "               [-O] [-s <img_size>] [ -S <swap_size> ] <img> <arch>"
  echo
  prose "Produce preconfigured parabola GNU/Linux-libre virtual machine instances."
  echo
  prose "The produced image file is written to <img>, and is configured and
         bootstrapped for the achitecture specified in <arch>. <arch> can be any
         one of the supported architectures: 'x86_64', 'i686' or 'armv7h',
         or one of the experimental arches: 'ppc64le' or 'riscv64'."
  echo
  echo  "Supported options:"
  echo  "  -h              Display this help and exit"
  echo  "  -H <hook>       Enable a hook to customize the created image. This can be"
  echo  "                  the path to a script, which will be executed once within"
  echo  "                  the running VM, or one of the predefined hooks described"
  echo  "  -k <kernel>     Choose the kernel package (default: $DEF_KERNEL)"
  echo  "                  below. This option can be specified multiple times"
  echo  "  -M <mirror>     Choose a different mirror from which to fetch packages"
  echo  "                  (default: $DEF_MIRROR)"
  echo  "  -O              Bootstrap an openrc system instead of a systemd one"
  echo  "  -s <img_size>   Set the size (in GB) of the VM image (default: $DEF_GB)"
  echo  "  -S <swap_size>  Set the size (in MB) of the swap partition (default: $DEF_SWAP_MB)"
  echo
  echo  "Predefined hooks:"
  echo  "  ethernet-dhcp:  configure and enable an ethernet device in the virtual"
  echo  "                  machine, using openresolv, dhcpcd and systemd-networkd"
  echo
  echo  "This script is part of parabola-vmbootstrap. source code available at:"
  echo  " <https://git.parabola.nu/~oaken-source/parabola-vmbootstrap.git>"
}

pvm_native_arch() {
  local arch=$( [[ "$1" =~ arm.* ]] && echo 'armv7l' || echo "$1" )

  setarch "$arch" /bin/true 2>/dev/null || return "$EXIT_FAILURE"
}

pvm_bootstrap() {
  msg "%s: starting image creation for %s" "$file" "$arch"

  # create the raw image file
  qemu-img create -f raw "$file" "${ImgSizeGb}G" || return "$EXIT_FAILURE"

  # prepare for cleanup
  trap 'pvm_cleanup' INT TERM RETURN

  # mount the virtual disk
  local workdir loopdev
  workdir="$(mktemp -d -t pvm-rootfs-XXXXXXXXXX)"  || return "$EXIT_FAILURE"
  loopdev="$(sudo losetup -fLP --show "$file")"    || return "$EXIT_FAILURE"
  sudo dd if=/dev/zero of="$loopdev" bs=1M count=8 || return "$EXIT_FAILURE"

  # partition
  local boot_begin="$( [[ "$arch" =~ i686|x86_64 ]] && echo 2 || echo 1 )MiB"
  local boot_end="$(( ${boot_begin/MiB} + $BootSizeMb ))MiB"
  local swap_begin="${boot_end}MiB"
  local swap_end="$(( ${swap_begin/MiB} + $SwapSizeMb ))MiB"
  local root_begin="${swap_end}MiB"
  local legacy_part boot_label boot_fs_type boot_flag
  case "$arch" in
    i686|x86_64)
        legacy_part='mkpart primary 1MiB 2Mib'
        boot_label='primary'
        boot_fs_type='ext2'
        boot_flag='bios_grub' # legacy_part
    armv7h)
        boot_label='ESP'
        boot_fs_type='fat32'
        boot_flag='boot'
    ppc64le|riscv64)
        boot_label='primary'
        boot_fs_type='ext2'
        boot_flag='boot'
  esac
  local swap_label='primary'
  local root_label='primary'
  sudo parted -s "$loopdev"                                \
    mklabel gpt                                            \
    $legacy_part                                           \
    mkpart $boot_label $boot_fs_type $boot_begin $boot_end \
    set 1 $boot_flag on                                    \
    mkpart $swap_label linux-swap $swap_begin $swap_end    \
    mkpart $root_label ext4 $root_begin 100%               || return "$EXIT_FAILURE"

  # refresh partition data
  sudo partprobe "$loopdev"

  # make file systems
  local boot_mkfs_cmd boot_loopdev swap_loopdev root_loopdev
  case "$arch" in
    i686|x86_64)
      boot_mkfs_cmd='mkfs.ext2'
      boot_loopdev="$loopdev"p2
      swap_loopdev="$loopdev"p3
      root_loopdev="$loopdev"p$( (( $SwapSizeMb > 0 )) && echo 4 || echo 3 )
      ;;
    armv7h)
      boot_mkfs_cmd='mkfs.vfat -F 32'
      boot_loopdev="$loopdev"p1
      swap_loopdev="$loopdev"p2
      root_loopdev="$loopdev"p$( (( $SwapSizeMb > 0 )) && echo 3 || echo 2 )
      ;;
    ppc64le|riscv64)
      boot_mkfs_cmd='mkfs.ext2'
      boot_loopdev="$loopdev"p1
      swap_loopdev="$loopdev"p2
      root_loopdev="$loopdev"p$( (( $SwapSizeMb > 0 )) && echo 3 || echo 2 )
      ;;
  esac
  sudo $boot_mkfs_cmd "$boot_loopdev" || return "$EXIT_FAILURE"
  sudo mkswap         "$swap_loopdev" || return "$EXIT_FAILURE"
  sudo mkfs.ext4      "$root_loopdev" || return "$EXIT_FAILURE"

  # mount partitions
  sudo mount "$root_loopdev" "$workdir"      || return "$EXIT_FAILURE"
  sudo mkdir -p              "$workdir"/boot || return "$EXIT_FAILURE"
  sudo mount "$boot_loopdev" "$workdir"/boot || return "$EXIT_FAILURE"

  # setup qemu-user-static, if necessary
  if ! pvm_native_arch "$arch"; then
    # target arch can't execute natively, pacstrap is going to need help by qemu
    local qemu_arch
    case "$arch" in
      armv7h) qemu_arch=arm ;;
      *) qemu_arch="$arch" ;;
    esac

    local qemu_user_static=$(sudo grep -l -F -e "interpreter /usr/bin/qemu-$qemu_arch-"   \
                                             -r -- /proc/sys/fs/binfmt_misc 2>/dev/null | \
                                  xargs -r sudo grep -xF 'enabled'                        )
    if [[ -z "$qemu_user_static" ]]; then
      error "%s: missing qemu-user-static for %s" "$file" "$arch"
      return "$EXIT_FAILURE"
    fi

    sudo mkdir -p "$workdir"/usr/bin
    sudo cp -v "/usr/bin/qemu-$qemu_arch-"* "$workdir"/usr/bin || return "$EXIT_FAILURE"
  fi

  # prepare pacstrap config
  local pacconf
  pacconf="$(mktemp -t pvm-pacconf-XXXXXXXXXX)" || return "$EXIT_FAILURE"
  cat > "$pacconf" << EOF
[options]
Architecture = $arch
$( (( $IsNonsystemd )) && echo -e "[nonsystemd]\nServer = $Mirror")
[libre]
Server = $Mirror
[core]
Server = $Mirror
[extra]
Server = $Mirror
[community]
Server = $Mirror
[pcr]
Server = $Mirror
EOF

  # prepare lists of packages
  local pkgs=(base $Kernel openssh openresolv ldns mc)
  [[ "${Hooks[@]}" =~ 'hook-ethernet-dhcp.sh' ]] && pkgs+=(dhcpcd netctl)
  case "$arch" in
    i686|x86_64) pkgs+=(grub) ;;
  esac
  case "$arch" in
    riscv64) ;;
    *) pkgs+=("haveged" net-tools) ;;
  esac
  local pkg_guest_cache=(ca-certificates-utils)

  # pacstrap! :)
  sudo pacstrap -GMc -C "$pacconf" "$workdir" "${pkgs[@]}"            || return "$EXIT_FAILURE"
  sudo pacstrap -GM  -C "$pacconf" "$workdir" "${pkg_guest_cache[@]}" || return "$EXIT_FAILURE"

  # create an fstab
  case "$arch" in
    riscv64) ;;
    *)
      if (( $SwapSizeMb > 0 )) ; then sudo swapoff --all ; sudo swapon "$swap_loopdev" ; fi ;
      genfstab -U "$workdir" | sudo tee "$workdir"/etc/fstab
      if (( $SwapSizeMb > 0 )) ; then sudo swapoff "$swap_loopdev" ; sudo swapon --all ; fi ;
      ;;
  esac

  # produce a hostname
  local hostname='parabola'
  echo -n "hostname: " ; echo $hostname | sudo tee "$workdir"/etc/hostname

  # produce an /etc/locale.conf
  local lang='en_US.UTF-8'
  echo -n "lang: " ; echo "LANG=$lang" | sudo tee "$workdir"/etc/locale.conf
  sudo sed -i "s/#${lang}/${lang}/" "$workdir"/etc/locale.gen

  # install a boot loader
  case "$arch" in
    i686|x86_64)
      local grub_def_file="$workdir"/etc/default/grub
      local grub_cfg_file=/boot/grub/grub.cfg
      # enable serial console
      local field=GRUB_CMDLINE_LINUX_DEFAULT
      local value="console=tty0 console=ttyS0"
      sudo sed -i "s/.*$field=.*/$field=\"$value\"/" "$grub_def_file" || return "$EXIT_FAILURE"
      # disable boot menu timeout
      local field=GRUB_TIMEOUT
      local value=0
      sudo sed -i "s/.*$field=.*/$field=$value/" "$grub_def_file"     || return "$EXIT_FAILURE"
      # install grub to the VM
      sudo arch-chroot "$workdir" grub-install "$loopdev"             || return "$EXIT_FAILURE"
      sudo arch-chroot "$workdir" grub-mkconfig -o $grub_cfg_file     || return "$EXIT_FAILURE"
      ;;
    riscv64)
      # FIXME: for the time being, use fedora bbl to boot
      local bbl_url=https://fedorapeople.org/groups/risc-v/disk-images/bbl
      sudo wget $bbl_url -O "$workdir"/boot/bbl || return "$EXIT_FAILURE"
      ;;
    # armv7h has no boot loader.
    # FIXME: what about ppc64le
  esac

  # regenerate the initcpio, skipping the autodetect hook
  local preset_file="$workdir"/etc/mkinitcpio.d/${Kernel}.preset
  sudo cp "${preset_file}"{,.backup}                                      || return "$EXIT_FAILURE"
  echo "default_options=\"-S autodetect\"" | sudo tee -a "${preset_file}" || return "$EXIT_FAILURE"
  sudo arch-chroot "$workdir" mkinitcpio -p ${Kernel}                     || return "$EXIT_FAILURE"
  sudo mv "${preset_file}"{.backup,}                                      || return "$EXIT_FAILURE"

  # disable audit
  sudo arch-chroot "$workdir" systemctl mask systemd-journald-audit.socket

  # initialize the pacman keyring
  sudo arch-chroot "$workdir" pacman-key --init
  sudo arch-chroot "$workdir" pacman-key --populate archlinux archlinux32 archlinuxarm parabola

  # enable the entropy daemon, to avoid stalling https
  sudo arch-chroot "$workdir" systemctl enable haveged.service

  # push hooks into the image
  sudo mkdir -p "$workdir/root/hooks"
  [ "${#Hooks[@]}" -eq 0 ] || sudo cp -v "${Hooks[@]}" "$workdir"/root/hooks/

  # create a master hook script
  sudo tee "$workdir"/root/hooks.sh << 'EOF'
#!/bin/bash
echo "[VM]: boot successful - configuring ...."

systemctl disable preinit.service

# generate the locale
locale-gen

# fix the mkinitcpio
mkinitcpio -p ${Kernel}

# fix ca-certificates
pacman -U --noconfirm /var/cache/pacman/pkg/ca-certificates-utils-*.pkg.tar.xz

# run the hooks
for hook in /root/hooks/*; do
  echo "[VM]: running hook \"$hook\""
  source "$hook" || return
done

# clean up after yourself
rm -rf /root/hooks
rm -f /root/hooks.sh
rm -f /usr/lib/systemd/system/preinit.service
rm -f /var/cache/pacman/pkg/*
rm -f /root/.bash_history

# report success :)
echo "[VM]: preinit hooks successful - powering off" ;
EOF

  # create a preinit service to run the hooks
  sudo tee "$workdir"/usr/lib/systemd/system/preinit.service << 'EOF'
[Unit]
Description=Oneshot VM Preinit
After=multi-user.target

[Service]
StandardOutput=journal+console
StandardError=journal+console
ExecStart=/usr/bin/bash /root/hooks.sh
Type=oneshot
ExecStopPost=shutdown -r now

[Install]
WantedBy=multi-user.target
EOF

  # enable the preinit service
  sudo arch-chroot "$workdir" systemctl enable preinit.service || return "$EXIT_FAILURE"

  # unmount everything
  pvm_cleanup

  # boot the machine to run the preinit hooks
  local qemu_flags=(-no-reboot)
  local pvmboot

  if [ -f "$THIS_DIR/pvmboot.sh" ]; then # in-tree
    pvmboot=(bash "$THIS_DIR/pvmboot.sh")
  elif type -p pvmboot &>/dev/null; then # installed
    pvmboot=(pvmboot)
  else
    error "%s: pvmboot not available -- unable to run hooks" "$file"
    return "$EXIT_FAILURE"
  fi

  exec 3>&1
  msg "booting the machine to run the preinit hooks"
  local vm_cmd="DISPLAY='' \"${pvmboot[@]}\" \"$file\" \"${qemu_flags[@]}\""
  $vm_cmd | tee /dev/fd/3 | grep ' [VM]: '| grep -q "[VM]: preinit hooks successful"
  local res=$?
  exec 3>&-

  if [ "$res" -ne 0 ]; then
    error "%s: failed to complete preinit hooks" "$file"
    return "$res"
  fi
}

pvm_cleanup() {
  trap - INT TERM RETURN

  echo "cleaning up"
  [ -n "$pacconf" ] && rm -f "$pacconf"
  unset pacconf
  if [ -n "$workdir" ]; then
    sudo rm -f "$workdir"/usr/bin/qemu-*
    sudo umount -R "$workdir" 2> /dev/null
    rmdir "$workdir"
  fi
  unset workdir
  [ -n "$loopdev" ] && sudo losetup -d "$loopdev"
  unset loopdev
}

main() {
  if [ "$(id -u)" -eq 0 ]; then
    error "This program must be run as a regular user"
    exit "$EXIT_NOPERMISSION"
  fi

  # parse options
  while getopts 'hH:kM:Os:' arg; do
    case "$arg" in
      h) usage; return "$EXIT_SUCCESS";;
      H) if [ -e   "$THIS_DIR/hooks/hook-$OPTARG.sh" ]; then                  # in-tree
           Hooks+=("$THIS_DIR/hooks/hook-$OPTARG.sh")
         elif [ -e "/usr/lib/libretools/pvmbootstrap/hook-$OPTARG.sh" ]; then # installed
           Hooks+=("/usr/lib/libretools/pvmbootstrap/hook-$OPTARG.sh")
         elif [ -e "$OPTARG" ]; then
           Hooks+=("$OPTARG")
         else
           warning "%s: hook does not exist" "$OPTARG"
         fi ;;
      k) Kernel="$OPTARG";;
      M) Mirror="$OPTARG";;
      O) IsNonsystemd=1;;
      s) ImgSizeGb="$( sed 's|[^0-9]||g' <<<$OPTARG)";;
      S) SwapSizeMb="$(sed 's|[^0-9]||g' <<<$OPTARG)";;
      *) usage >&2; exit "$EXIT_INVALIDARGUMENT";;
    esac
  done

  local shiftlen=$(( OPTIND - 1 ))
  shift $shiftlen
  local file="$1"
  local arch="$2"
  local has_params=$( [ "$#" -eq 2 ] && echo 1 || echo 0 )
  local has_space=$( (( ($ImgSizeGb*1000) > ($MIN_GB*1000) + $SwapSizeMb )) && echo 1 || echo 0 )
  (( ! $has_params )) || (( ! $has_space )) && usage >&2 && exit "$EXIT_INVALIDARGUMENT"

  # determine if the target arch is supported
  case "$arch" in
    i686|x86_64|armv7h) ;;
    ppc64le|riscv64)
      warning "%s: arch %s is experimental" "$file" "$arch";;
    *)
      error "%s: arch %s is unsupported" "$file" "$arch"
      exit "$EXIT_INVALIDARGUMENT";;
  esac

  # determine whether the target output file already exists
  if [ -e "$file" ]; then
    warning "%s: file exists. Continue? [y/N]" "$file"
    read -p " " -n 1 -r
    echo
    if [[ ! $REPLY =~ ^[Yy]$ ]]; then
      exit "$EXIT_FAILURE"
    fi
    rm -f "$file" || exit
  fi

  # create the virtual machine
  if ! pvm_bootstrap; then
    error "%s: bootstrap failed" "$file"
    exit "$EXIT_FAILURE"
  fi

  msg "%s: bootstrap complete" "$file"
}

main "$@"
